package org.processmining.ptandloggenerator.algorithms;

import java.util.Date;
import java.util.Random;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.ptandloggenerator.parameters.NoiseGeneratorSettings;

/**
 * Utility class with functions to add noise to an event log on a trace level.
 * 
 * @author abolt
 *
 */
public class NoiseGenerator {

	NoiseGeneratorSettings settings;
	XLog log;

	public NoiseGenerator(NoiseGeneratorSettings settings, XLog log) {
		this.settings = settings;
		this.log = log;
	}

	public XTrace addNoiseToTrace(XTrace originalTrace) {

		//first we create a clone of the trace that can be modified by any sub-function
		XTrace newTrace = (XTrace) originalTrace.clone();

		Random random = new Random();

		//check if we have to enforce a single noise operation
		if (settings.getForceSingleNoiseOp()) {
			boolean isDone = false;

			while (!isDone) {
				switch (random.nextInt(8)) {
					case 0 :
						if (settings.getProbRemoveBody() > 0 && newTrace.size() > 3) {
							removeBody(newTrace);
							isDone = true;
						}
						break;
					case 1 :
						if (settings.getProbRemoveEvent() > 0 && newTrace.size() > 1) {
							removeEvent(newTrace);
							isDone = true;
						}
						break;
					case 2 :
						if (settings.getProbRemoveHead() > 0 && newTrace.size() > 3) {
							removeHead(newTrace);
							isDone = true;
						}
						break;
					case 3 :
						if (settings.getProbRemoveTail() > 0 && newTrace.size() > 3) {
							removeTail(newTrace);
							isDone = true;
						}
						break;
					case 4 :
						if (settings.getProbSwapConsecutiveEvents() > 0 && newTrace.size() > 1) {
							swapConsecutive(newTrace);
							isDone = true;
						}
						break;
					case 5 :
						if (settings.getProbSwapRandomEvents() > 0 && newTrace.size() > 1) {
							swapRandom(newTrace);
							isDone = true;
						}
						break;
					case 6 :
						if (settings.getProbAddEvent() > 0) {
							addEvent(newTrace);
							isDone = true;
						}
						break;
					case 7 :
						if (settings.getProbDuplicateEvent() > 0) {
							duplicateEvent(newTrace);
							isDone = true;
						}
						break;
				}
			}
		} else {
			//add an event (existing in the log) in a random position
			if (random.nextDouble() < settings.getProbAddEvent())
				addEvent(newTrace);

			if (random.nextDouble() < settings.getProbDuplicateEvent() && newTrace.size() > 0)
				duplicateEvent(newTrace);

			if (random.nextDouble() < settings.getProbRemoveBody() && newTrace.size() > 2)
				removeBody(newTrace);

			if (random.nextDouble() < settings.getProbRemoveEvent() && newTrace.size() > 1)
				removeEvent(newTrace);

			if (random.nextDouble() < settings.getProbRemoveHead() && newTrace.size() > 2)
				removeHead(newTrace);

			if (random.nextDouble() < settings.getProbRemoveTail() && newTrace.size() > 2)
				removeTail(newTrace);

			if (random.nextDouble() < settings.getProbSwapConsecutiveEvents() && newTrace.size() > 1)
				swapConsecutive(newTrace);

			if (random.nextDouble() < settings.getProbSwapRandomEvents() && newTrace.size() > 1)
				swapRandom(newTrace);
		}
		return newTrace;
	}

	private void addEvent(XTrace trace) {
		Random random = new Random();
		int duplicateIndex = random.nextInt(trace.size());
		XTrace oldTrace = log.get(random.nextInt(log.size()));
		XEvent oldEvent = oldTrace.get(random.nextInt(oldTrace.size()));

		XEvent event = (XEvent) oldEvent.clone();
		Date time = XTimeExtension.instance().extractTimestamp(trace.get(duplicateIndex));
		if (time != null)
			XTimeExtension.instance().assignTimestamp(event, time.getTime() - 1);
		trace.add(duplicateIndex, event);
	}

	private void duplicateEvent(XTrace trace) {
		Random random = new Random();
		int duplicateIndex = random.nextInt(trace.size());
		XEvent oldEvent = trace.get(duplicateIndex);
		XEvent event = (XEvent) oldEvent.clone();
		Date time = XTimeExtension.instance().extractTimestamp(oldEvent);
		if (time != null)
			XTimeExtension.instance().assignTimestamp(event, time.getTime() - 1);
		trace.add(duplicateIndex, event);
	}

	//removes up to 1/3 of the first events of a trace 
	private void removeHead(XTrace trace) {
		Random random = new Random();
		int headIndex = random.nextInt(trace.size() / 3);
		for (int i = 0; i <= headIndex; i++)
			//removes always the first event since the list is updated
			trace.remove(0);
	}

	//removes up to 1/3 of the last events of a trace 
	private void removeTail(XTrace trace) {
		Random random = new Random();
		int windowSize = random.nextInt(trace.size() / 3);
		int tailIndex = trace.size() - windowSize - 1;
		for (int i = 0; i <= windowSize; i++)
			trace.remove(tailIndex);
	}

	//removes up to 1/3 of the middle events of a trace (starting from a random index)
	private void removeBody(XTrace trace) {
		Random random = new Random();
		int windowSize = random.nextInt(trace.size() / 3);
		int bodyIndex = random.nextInt(trace.size() - windowSize - 1);
		for (int i = 0; i <= windowSize; i++)
			trace.remove(bodyIndex);
	}

	//remove an event in a random index
	private void removeEvent(XTrace trace) {
		Random random = new Random();
		trace.remove(random.nextInt(trace.size()));
	}

	private void swapConsecutive(XTrace trace) {
		Random random = new Random();
		//we do not swap the last one
		int index1 = random.nextInt(trace.size() - 1);
		int index2 = index1 + 1;
		Date time1 = XTimeExtension.instance().extractTimestamp(trace.get(index1));
		Date time2 = XTimeExtension.instance().extractTimestamp(trace.get(index2));

		XEvent event1 = trace.get(index1);
		XEvent event2 = trace.get(index2);
		//swap timestamps first

		if (time2 != null)
			XTimeExtension.instance().assignTimestamp(event1, time2);
		if (time1 != null)
			XTimeExtension.instance().assignTimestamp(event2, time1);
		//now swap the events
		trace.set(index1, event2);
		trace.set(index2, event1);
	}

	private void swapRandom(XTrace trace) {

		Random random = new Random();
		//we do not swap the last one
		int index1 = random.nextInt(trace.size());
		int index2;

		do {
			index2 = random.nextInt(trace.size());
		} while (index1 == index2);

		Date time1 = XTimeExtension.instance().extractTimestamp(trace.get(index1));
		Date time2 = XTimeExtension.instance().extractTimestamp(trace.get(index2));

		XEvent event1 = trace.get(index1);
		XEvent event2 = trace.get(index2);
		//swap timestamps first
		if (time2 != null)
			XTimeExtension.instance().assignTimestamp(event1, time2);
		if (time1 != null)
			XTimeExtension.instance().assignTimestamp(event2, time1);
		//now swap the events
		trace.set(index1, event2);
		trace.set(index2, event1);
	}
}
