package org.processmining.ptandloggenerator.plugins;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.swing.JOptionPane;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.balancedconformance.BalancedDataXAlignmentPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverNoOpImpl;
import org.processmining.plugins.petrinet.finalmarkingprovider.MarkingEditorPanel;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.processmining.ptandloggenerator.algorithms.NoiseGenerator;
import org.processmining.ptandloggenerator.parameters.NoiseGeneratorSettings;

@Plugin(name = "Add Noise to Logg", parameterLabels = { "Log", "Petri Net", "Initial Marking", "Final Marking",
		"Classifier" }, returnLabels = { "XLog" }, returnTypes = {
				XLog.class }, userAccessible = true, help = "Add noise to an event log. Each trace is guaranteed to not fit the process model given as input")

public class Alfredo_GenerateNoisyLog {

	@UITopiaVariant(affiliation = "Eindhoven University of Technology", author = "Alfredo Bolt", email = "a.bolt@tue.nl")
	@PluginVariant(variantLabel = "Generate a Noisy Log using default Parameters", requiredParameterLabels = { 0, 1 })
	public XLog addNoiseWithDefaultParameters(UIPluginContext context, XLog log, Petrinet model) {

		Marking initMarking = null;
		try {
			initMarking = context.getConnectionManager()
					.getFirstConnection(InitialMarkingConnection.class, context, model)
					.getObjectWithRole(FinalMarkingConnection.MARKING);
			if (initMarking.isEmpty()) {
				int yn = JOptionPane.showConfirmDialog(null, "The initial marking is empty. Is this intended?",
						"Initial Marking", JOptionPane.YES_NO_OPTION);
				if (yn == JOptionPane.NO_OPTION) {
					JOptionPane.showMessageDialog(null,
							"Please create the correct initial marking using the right plug-in");
					return null;
				}
			}
		} catch (Exception e) {
			initMarking = createMarking(context, model, "Initial Marking");
			ProvidedObjectHelper.publish(context, "Initial Marking for " + model.getLabel(), initMarking, Marking.class,
					false);
			context.getConnectionManager().addConnection(new InitialMarkingConnection(model, initMarking));
		}

		Marking finalMarking = null;
		try {
			finalMarking = context.getConnectionManager()
					.getFirstConnection(FinalMarkingConnection.class, context, model)
					.getObjectWithRole(FinalMarkingConnection.MARKING);
			if (finalMarking.isEmpty()) {
				int yn = JOptionPane.showConfirmDialog(null, "The final marking is empty. Is this intended?",
						"Final Marking", JOptionPane.YES_NO_OPTION);
				if (yn == JOptionPane.NO_OPTION) {
					JOptionPane.showMessageDialog(null,
							"Please create the correct final marking using the right plug-in");
					return null;
				}
			}

		} catch (Exception e) {
			finalMarking = createMarking(context, model, "Final Marking");
			ProvidedObjectHelper.publish(context, "Final Marking for " + model.getLabel(), finalMarking, Marking.class,
					false);
			context.getConnectionManager().addConnection(new FinalMarkingConnection(model, finalMarking));
		}

		return addNoise(context, log, model, initMarking, finalMarking, XLogInfoImpl.NAME_CLASSIFIER, null);
	}

	@UITopiaVariant(affiliation = "Eindhoven University of Technology", author = "Alfredo Bolt", email = "a.bolt@tue.nl")
	@PluginVariant(variantLabel = "Generate a Noisy Log", requiredParameterLabels = { 0, 1, 2, 3, 4 })
	public XLog addNoise(PluginContext context, XLog log, Petrinet net, Marking initialMarking, Marking finMarking,
			XEventClassifier classifier, NoiseGeneratorSettings settings) {

		XLog result = (XLog) log.clone();
		result.clear();
		XLog auxLog = (XLog) log.clone();

		final Map<Integer, Boolean> nonFittingTraces = new HashMap<Integer, Boolean>();
		for (int i = 0; i < auxLog.size(); i++)
			nonFittingTraces.put(i, false);

		final List<XTrace> traceList = new ArrayList<XTrace>();

		BalancedDataXAlignmentPlugin balanced = new BalancedDataXAlignmentPlugin();
		if (settings == null)
			settings = NoiseGeneratorSettings.getDefaultSettings();

		NoiseGenerator generator = new NoiseGenerator(settings, auxLog);

		Marking[] finalMarking = new Marking[1];
		finalMarking[0] = finMarking;
		BalancedProcessorConfiguration conformanceSettings = BalancedProcessorConfiguration.newDefaultInstance(net,
				initialMarking, finalMarking, auxLog, classifier, 1, 1, 1, 1);
		conformanceSettings.setMaxCostFactor(1);
		conformanceSettings.setObserver(new DataConformancePlusObserverNoOpImpl() {

			public void foundImpossibleAlignments(Collection<ImpossibleTrace> impossibleTraces) {
				for (ImpossibleTrace trace : impossibleTraces)
					traceList.add(trace.getTrace());
			}
		});

		try {
			int counter = 0;
			int modLimit = 5; // max noise runs made to a trace
			Random random = new Random();
			do {
				//extract the traces that do not fit
				for (XTrace trace : traceList) {
					result.add(trace);
					auxLog.remove(trace);
				}
				traceList.clear();

				//reduce the log, keeping only those who still fit, add noise to the fitting traces
				for (int i = 0; i < auxLog.size(); i++) {
					if (counter < modLimit)
						auxLog.set(i, generator.addNoiseToTrace(auxLog.get(i)));
					else {
						//if after X tries we still cannot break the trace, we take a random trace from the original log.
						auxLog.set(i, generator.addNoiseToTrace(log.get(random.nextInt(log.size()))));
					}
				}
				//reset counter
				if (counter >= modLimit){
					counter = 0;
					//modLimit++;
				}

				balanced.alignLogPluginConvertFromPetrinet(context, net, auxLog, conformanceSettings);
				counter++;
			} while (!auxLog.isEmpty());
			return result;

		} catch (ControlFlowAlignmentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (DataAlignmentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return null;
	}

	private static Marking createMarking(UIPluginContext context, PetrinetGraph net, String title) {
		MarkingEditorPanel editor = new MarkingEditorPanel(title);

		return editor.getMarking(context, net);
	}
}
