package org.processmining.ptandloggenerator.plugins;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.ptandloggenerator.algorithms.LogCollectionGenerator;
import org.processmining.ptandloggenerator.algorithms.NoiseGenerator;
import org.processmining.ptandloggenerator.dialogs.logSettingsNoise;
import org.processmining.ptandloggenerator.models.LogCollection;
import org.processmining.ptandloggenerator.models.NewickTreeCollection;
import org.processmining.ptandloggenerator.parameters.NoiseGeneratorSettings;

import cern.colt.Arrays;

@Plugin(name = "Generate Log Collection (with noise) from Process Trees", parameterLabels = { "Newick Tree Collection", "Log Generator Parameters" }, returnLabels = {
"Log collection (XES format)" }, returnTypes = { LogCollection.class }, 
help = "Generate events logs for a given sample of process trees. URI: http://hdl.handle.net/1942/20818")
public class GenerateLogCollectionNoise {
	
	@UITopiaVariant(affiliation = "Hasselt University", author = "Toon Jouck", email = "toon.jouck@uhasselt.be")
	@PluginVariant(variantLabel = "Generate (noisy) Event Logs (XES) from Process Trees (Newick Trees)", requiredParameterLabels = { 0 })
	public LogCollection run(UIPluginContext context, NewickTreeCollection collection) {
		context.log("Creating log settings");
		//set maximum for progressbar
		context.getProgress().setMaximum(collection.getNewickTreeList().size());
		context.getProgress().setValue(0);
		
		//initialize settings object
		logSettingsNoise settings = new logSettingsNoise();
		
		//show input wizard
		InteractionResult result = context.showWizard("Log generator Settings", true, true, settings.initComponents());
		
		switch (result) {
			case CANCEL :
				return cancel(context, collection, "The user has cancelled the Log generator!");
			case FINISHED :
				settings.setNoCases();
				settings.setNoisePercentage();
				context.log("Starting log generation");
				LogCollectionGenerator generator = new LogCollectionGenerator(collection, settings.getNoCases());
				
				int i = 1;
				
				for (XLog log: generator.getLogs().getXLogCollection()){
					
					System.out.println("LOG " + i);
					
					BigDecimal zeroNoise = new BigDecimal(0.0);
					XLog auxLog = (XLog) log.clone();
					
					if (settings.getNoisePercentage().compareTo(zeroNoise) != 0) {
						// add noise
						
						// first fix the settings of the noise generator
						NoiseGeneratorSettings noiseSettings = new NoiseGeneratorSettings();
						noiseSettings = NoiseGeneratorSettings.getIntegratedDefaultSettings();
						
						//initialize the noise generator object
						NoiseGenerator noiseGenerator = new NoiseGenerator(noiseSettings, log);
						int numberOfNoisyTraces = settings.getNoisePercentage().multiply(new BigDecimal(settings.getNoCases())).intValue();
						
						//test if the number of noisy traces does not exceed the number of traces with more than one event
						if (numberOfNoisyTraces > (auxLog.size() - numberOfSingleEventTraces(auxLog))){
							numberOfNoisyTraces = auxLog.size() - numberOfSingleEventTraces(auxLog);
						}
						List<Integer> usedIndexes = new ArrayList<Integer>();
						Random random = new Random();
						int randomIndex = random.nextInt(log.size());
						
						for (int j = 0; j < numberOfNoisyTraces; j++){
							
							while (usedIndexes.contains(randomIndex) || log.get(randomIndex).size() < 2) {
							randomIndex = random.nextInt(log.size());
							}
							auxLog.set(randomIndex, noiseGenerator.addNoiseToTrace(log.get(randomIndex)));
							usedIndexes.add(randomIndex);
						}
						
						System.out.println(Arrays.toString(usedIndexes.toArray()));
						
					}
					//unpack created log
					context.getProvidedObjectManager().createProvidedObject("Log " + Integer.toString(i), auxLog, XLog.class , context);
					i++;
					org.processmining.plugins.utils.ProvidedObjectHelper.setFavorite(context, auxLog);
					
					//advance the progressbar
					context.getProgress().inc();
				}
				return generator.getLogs();
			default :
				System.out.println("problem with the settings");
				return null;
			
		}

	}
	
	private static int numberOfSingleEventTraces(XLog log){
		int singleEventTraces = 0;
		for (XTrace trace: log){
			if (trace.size() == 1){
				singleEventTraces++;
			}
		}
		return singleEventTraces;
	}
	
	/**
	* Cancel the future results, display a error message and return all nulls.
	* @param context Context of the plug-in.
	* @param msg Error message.
	* @return Return an object array of nulls;
	*/
	private LogCollection cancel(PluginContext context, NewickTreeCollection collection, String msg) {
		context.log(msg);
		context.getFutureResult(0).cancel(true);
		context.getFutureResult(1).cancel(true);
		return new LogCollectionGenerator(collection, 0).getLogs();
	}

}
