package org.processmining.ptandloggenerator.plugins;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.ptandloggenerator.algorithms.TreeFactory;
import org.processmining.ptandloggenerator.dialogs.treeSettings;
import org.processmining.ptandloggenerator.models.NewickTree;
import org.processmining.ptandloggenerator.models.NewickTreeCollection;


@Plugin(name = "Generate Process Trees from Population", parameterLabels = { }, 
returnLabels = { "Newick Tree Collection (Process Trees)" }, returnTypes = { NewickTreeCollection.class },
help = "Define a population of process trees from which a random sample of chosen size will be returned. URI: http://hdl.handle.net/1942/20818")
public class GenerateTreeUI {
	
	/*
	 * Plugin that generates a Newick Tree Collection based on population specifications from UI (using a wizard)
	 */
	
	//Initialise factory
	private TreeFactory factory = null;
	public NewickTreeCollection results;
	
	@UITopiaVariant(affiliation = "Hasselt University", author = "Toon Jouck", email = "toon.jouck@uhasselt.be")
	@PluginVariant(variantLabel = "Generate NewickTree (Process Tree) from user defined population", requiredParameterLabels = { })
	public NewickTreeCollection run(UIPluginContext context){
		//log initialise message
		context.log("Creating population settings");
		//Initialise settings object
		treeSettings settings = new treeSettings();
		
		//show input wizard
		InteractionResult result = context.showWizard("Tree Generator Settings", true, true, settings.initComponents());
		
		switch (result) {
			case CANCEL :
				return cancel(context, "The user has cancelled the Log generator!");
			case FINISHED :
				//finish settings
				settings.setAllSettings();
				
				//get progressbar
				context.getProgress().setMaximum(settings.getNoTrees());
				context.getProgress().setValue(0);
				
				//check if the probabilities operators sum to one
				if(settings.checkProbabilities()){
					
					//get parameters in right format
					String parameterline = Integer.toString(settings.getModeAct()) + ";"
							+ Integer.toString(settings.getMinAct()) + ";"
							+ Integer.toString(settings.getMaxAct()) + ";"
							+ (settings.getProbSeq().toString()) + ";"
							+ (settings.getProbCho().toString()) + ";"
							+ (settings.getProbPar().toString()) + ";"
							+ (settings.getProbLoop().toString()) + ";"
							+ (settings.getProbOr().toString()) + ";"
							+ (settings.getProbSilent().toString()) + ";"
							+ (settings.getProbDuplicate().toString()) + ";"
							+ (settings.getProbLTDependencies().toString()) + ";"
							+ (settings.getProbInfrequentPaths().toString()) + ";"
							+ Integer.toString(settings.getNoTrees()) + ";"
							+ settings.getUnfoldLoops() + ";"
							+ Integer.toString(settings.getLoopIterations()) + ";"
							//-1 denoting no random seed specified
							+ Integer.toString(-1);
					
					System.out.println(parameterline);
					
					context.log("Generating trees from population");
					factory = new TreeFactory();
					results = new NewickTreeCollection();
					
					for (int i = 0; i < settings.getNoTrees(); i++){
						results.addNewickTree(new NewickTree(
								factory.createTree(parameterline)));
						factory.cleanupInterpreter();
						context.getProgress().inc();
					}
					
					return results;
				}
				else
					return error(context, "User did not provide operator probabilities that sum to one!");
			default :
				System.out.println("problem with the settings");
				return null;
			
		}
		
	}
	
	/**
	* Cancel the future results, display a error message and return all nulls.
	* @param context Context of the plug-in.
	* @param msg Error message.
	* @return Return an object array of nulls;
	*/
	private NewickTreeCollection cancel(PluginContext context, String msg) {
		context.log(msg, MessageLevel.WARNING);
		context.getFutureResult(0).cancel(true);
		return null;
	}
	
	private NewickTreeCollection error(PluginContext context, String msg){
		context.log(msg,MessageLevel.ERROR);
		return new NewickTreeCollection();
	}

}
