package org.processmining.ptandloggenerator.plugins.data;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.impl.XLogInfoImpl;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.Alignment;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.BalancedDataConformancePlusPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration.UnassignedMode;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.controlflow.adapter.SearchMethod;
import org.processmining.plugins.balancedconformance.dataflow.DataAlignmentAdapter;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverNoOpImpl;
import org.processmining.plugins.balancedconformance.result.BalancedReplayResult;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.ptandloggenerator.algorithms.DataLogCollectionGenerator;
import org.processmining.ptandloggenerator.dialogs.dataEvaluationSettings;
import org.processmining.ptandloggenerator.models.DataLogCollection;
import org.processmining.ptandloggenerator.models.EvaluationResults;
import org.processmining.ptandloggenerator.models.EvaluationResultsCollection;
import org.processmining.ptandloggenerator.models.NewickTree;
import org.processmining.ptandloggenerator.models.NewickTreeCollection;
import org.processmining.ptandloggenerator.parsers.NewickToPetriNet;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;

import nl.tue.astar.AStarThread.ASynchronousMoveSorting;
import nl.tue.astar.AStarThread.QueueingModel;

@Plugin(name = "Decision mining evaluation (10-fold CV)", parameterLabels = { "Newick tree collection" }, returnLabels = {
"Evaluation results collection" }, returnTypes = { EvaluationResultsCollection.class }, 
help = "help")
public class DataEvaluationCV {
	
	@UITopiaVariant(affiliation = "Hasselt University", author = "Toon Jouck", email = "toon.jouck@uhasselt.be")
	@PluginVariant(variantLabel = "Evaluate decision mining algorithm", requiredParameterLabels = { 0 })
	public EvaluationResultsCollection evaluate(UIPluginContext context, NewickTreeCollection treeCollection) throws Exception {
		
		// set maximum for progressbar
		context.getProgress().setMaximum(treeCollection.getNewickTreeList().size());
		context.getProgress().setValue(0);
		
		// set data evaluation settings
		dataEvaluationSettings settings = new dataEvaluationSettings();
		
		//show input wizard
		InteractionResult interaction = context.showWizard("Decision Mining Evaluation Settings", true, true, settings.initComponents());
		
		switch (interaction) {
			case CANCEL :
				return cancel(context, treeCollection, "The user has cancelled the decision mining evaluation!");
			case FINISHED :
				settings.setMinNoCases();
				settings.setMaxNoCases();
				settings.setMaxInput();
				settings.setMaxCutoff();
				settings.setTargetDeterminism();
				settings.setMiner();
				// get random number of cases from a specified range
				int min = settings.getMinNoCases();
				int max = settings.getMaxNoCases();
				Random r = new Random();
				int num = r.nextInt(max-min) + min;
				// make sure the number of cases is divisible by 20
				int noCases = (num + 10)/20 *20;
				
				DataLogCollectionGenerator generator = new DataLogCollectionGenerator(treeCollection,
						Integer.toString(noCases),Integer.toString(noCases/2),settings.getMaxInput(),
						settings.getMaxCutoff(),settings.getTargetDeterminism());
				// add data logs to workspace
				context.getProvidedObjectManager().createProvidedObject("data log collection", 
						generator.getLogs(), DataLogCollection.class , context);
				org.processmining.plugins.utils.ProvidedObjectHelper.setFavorite(context, generator.getLogs());
				
				int i = 1;
				String miner;
				if (settings.getMiner())
					miner = "overlapping";
				else
					miner = "alignments";
				EvaluationResultsCollection results = new EvaluationResultsCollection(miner);
				List<NewickTree> treeList = treeCollection.getNewickTreeList();
				for (XLog log: generator.getLogs().getXLogCollection()){
					System.out.println("Model/log: " + i);
					// get tree and convert into Petri net
					NewickTree tree = treeList.get(i-1);
					Object[] result = NewickToPetriNet.convert(context, tree);
					
					// set log classifier
					log.getClassifiers().add(XLogInfoImpl.NAME_CLASSIFIER);
					
					// first split the log in fitting and non-fitting logs
					ArrayList<XLog> splitLogs = splitLog(context,log,noCases);
					
					// do 10-fold cross-validation
					for(int fold=1; fold<11; fold++) {
						//System.out.println("Validation fold: " + fold);
						ArrayList<XLog> logsCV = splitTrainingTest(context,splitLogs,fold);
						
						results.addToEvaluationResultsCollection(evaluate(context, (Petrinet) result[0],
								logsCV.get(0), logsCV.get(1), logsCV.get(2), i, fold, settings.getMiner()));
					}
					
					// update log index
					i++;
					//advance the progressbar
					context.getProgress().inc();
				}
				
				// Set appropriate name for returned results object
				context.getFutureResult(0).setLabel("Evaluation results");
				
				return results;
			default :
				System.out.println("problem with the settings");
				return null;
		}
	}
	
	private static ArrayList<XLog> splitLog(PluginContext context, XLog log, int noFittingCases) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog fitting = factory.createLog((XAttributeMap) log.getAttributes().clone());
		XLog nonfitting = factory.createLog((XAttributeMap) log.getAttributes().clone());
		fitting.getClassifiers().addAll(log.getClassifiers());
		fitting.getExtensions().addAll(log.getExtensions());
		fitting.getGlobalTraceAttributes().addAll(log.getGlobalTraceAttributes());
		fitting.getGlobalEventAttributes().addAll(log.getGlobalEventAttributes());
		nonfitting.getClassifiers().addAll(log.getClassifiers());
		nonfitting.getExtensions().addAll(log.getExtensions());
		nonfitting.getGlobalTraceAttributes().addAll(log.getGlobalTraceAttributes());
		nonfitting.getGlobalEventAttributes().addAll(log.getGlobalEventAttributes());
		for (XTrace trace : log) {
			for (String key : trace.getAttributes().keySet()) {
				String value = trace.getAttributes().get(key).toString();
				if (Integer.parseInt(value) < noFittingCases + 1) {
					fitting.add(trace);
				}
				else
					nonfitting.add(trace);
			}
		}
		XConceptExtension.instance().assignName(fitting, "fitting");
		XConceptExtension.instance().assignName(nonfitting, "non-fitting");
		ArrayList<XLog> logs = new ArrayList<XLog>();
		logs.add(fitting);
		logs.add(nonfitting);
		return logs;
	}
	
	private static ArrayList<XLog> splitTrainingTest(PluginContext context, ArrayList<XLog> logs, int run) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog training = factory.createLog((XAttributeMap) logs.get(0).getAttributes().clone());
		XLog testFitting = factory.createLog((XAttributeMap) logs.get(0).getAttributes().clone());
		XLog testNonFitting = factory.createLog((XAttributeMap) logs.get(0).getAttributes().clone());
		training.getClassifiers().addAll(logs.get(0).getClassifiers());
		training.getExtensions().addAll(logs.get(0).getExtensions());
		training.getGlobalTraceAttributes().addAll(logs.get(0).getGlobalTraceAttributes());
		training.getGlobalEventAttributes().addAll(logs.get(0).getGlobalEventAttributes());
		testFitting.getClassifiers().addAll(logs.get(1).getClassifiers());
		testFitting.getExtensions().addAll(logs.get(1).getExtensions());
		testFitting.getGlobalTraceAttributes().addAll(logs.get(1).getGlobalTraceAttributes());
		testFitting.getGlobalEventAttributes().addAll(logs.get(1).getGlobalEventAttributes());
		testNonFitting.getClassifiers().addAll(logs.get(1).getClassifiers());
		testNonFitting.getExtensions().addAll(logs.get(1).getExtensions());
		testNonFitting.getGlobalTraceAttributes().addAll(logs.get(1).getGlobalTraceAttributes());
		testNonFitting.getGlobalEventAttributes().addAll(logs.get(1).getGlobalEventAttributes());
		// add fitting traces
		int startIndex;
		int endIndex = (logs.get(0).size()/10)*run;
		if (run == 1)
			startIndex = 1;
		else
			startIndex = (logs.get(0).size()/10)*(run-1) + 1;
		int added = 0;
		// System.out.println("Fitting log: start index " + startIndex + ", end index " + endIndex);
		for (XTrace trace : logs.get(0)) {
			for (String key : trace.getAttributes().keySet()) {
				String value = trace.getAttributes().get(key).toString();
				if (Integer.parseInt(value) >= startIndex && Integer.parseInt(value) <= endIndex) {
					// only add first half of those traces:
					if (added < logs.get(0).size()/10/2) {
						testFitting.add(trace);
						added++;
						// System.out.println("added fitting test trace " + value);
					}
				}
				else {
					training.add(trace);
					// System.out.println("added training trace " + value);
				}
			}
		}
		// add non-fitting traces
		if (run == 1)
			startIndex = logs.get(0).size() + 1;
		else
			startIndex = logs.get(0).size() + logs.get(1).size()/10*(run-1) + 1;
		endIndex = logs.get(0).size() + (logs.get(1).size()/10)*run;
		// System.out.println("Non-fitting log: start index " + startIndex + ", end index " + endIndex);
		for (XTrace trace : logs.get(1)) {
			for (String key : trace.getAttributes().keySet()) {
				String value = trace.getAttributes().get(key).toString();
				if (Integer.parseInt(value) >= startIndex && Integer.parseInt(value) <= endIndex) {
					testNonFitting.add(trace);
					// System.out.println("added non-fitting test trace " + value);
				}
			}
		}
		ArrayList<XLog> logsCV = new ArrayList<XLog>();
		logsCV.add(training);
		logsCV.add(testFitting);
		logsCV.add(testNonFitting);
		return logsCV;
	}
	
	public static EvaluationResults evaluate(PluginContext context, Petrinet model, XLog trainingLog,
			XLog testLogFitting, XLog testLogNonFitting, int i, int fold, boolean useOverlapping)
			throws Exception, DataAlignmentException {
		// set log classifier
		trainingLog.getClassifiers().add(XLogInfoImpl.NAME_CLASSIFIER);
		
		// apply decision miner on training log
		DecisionMinerNoUI miner = new DecisionMinerNoUI();
		Object[] discoveryResults = new Object[3];
		discoveryResults = miner.decisionDiscovery(context, model, trainingLog, useOverlapping);
		// alternatively call decision miner with UI parameter settings
		// DecisionMining miner = new DecisionMining();
		// discoveryResults = miner.decisionMiner(context, model, splitLogs.get(0));
		
		Marking initialMarking = (Marking) discoveryResults[1];
		Marking[] finalMarking = new Marking[1];
		DataPetriNet dpn = (DataPetriNet) discoveryResults[0];
		finalMarking[0] = (Marking) discoveryResults[2];
		
		// fix configuration of balanced replayer
		BalancedProcessorConfiguration config = new BalancedProcessorConfiguration();
		config.setActivateDataViewCache(true);
		config.setUseOptimizations(true);
		config.setConcurrentThreads(Runtime.getRuntime().availableProcessors());
		config.setSorting(ASynchronousMoveSorting.LOGMOVEFIRST);
		config.setIlpSolver(DataAlignmentAdapter.ILPSolver.ILP_LPSOLVE);
		config.setUsePartialDataAlignments(false);
		config.setVariablesUnassignedMode(UnassignedMode.NULL);
		config.setQueueingModel(QueueingModel.DEPTHFIRST);
		config.setMaxCostFactor(0);
		config.setUsePartialOrders(false);

		config.setKeepControlFlowSearchSpace(true);
		config.setKeepDataFlowSearchSpace(true);
		config.setSearchMethod(SearchMethod.ASTAR_GRAPH);
		config.setMaxQueuedStates(100_000_001);
		
		// add training and test log to workspace
//		String trainingName = "Training log " + i + " fold " + fold;
//		String testFittingName = "Test log fitting " + i + " fold " + fold;
//		String testNonFittingName = "Test log non-fitting " + i + " fold " + fold;
//		context.getProvidedObjectManager().createProvidedObject(trainingName, 
//				trainingLog, XLog.class , context);
//		context.getProvidedObjectManager().createProvidedObject(testFittingName, 
//				testLogFitting, XLog.class , context);
//		context.getProvidedObjectManager().createProvidedObject(testNonFittingName, 
//				testLogNonFitting, XLog.class , context);
		
		// add discovered data petri net to workspace
		String dpnName = "Discovered data flow " + i + " fold " + fold;
		context.getProvidedObjectManager().createProvidedObject(dpnName, 
				dpn, DataPetriNet.class , context);
		// org.processmining.plugins.utils.ProvidedObjectHelper.setFavorite(context, dpn);
		
		// replay both fitting and non-fitting test logs
		BalancedReplayResult resultFitting = replay(context, config, testLogFitting, 
				dpn, initialMarking, finalMarking, i, fold);
		BalancedReplayResult resultNonFitting = replay(context, config, testLogNonFitting, 
				dpn, initialMarking, finalMarking, i, fold);
		
		// do alignments UI
		// BalancedReplayResult result = balanced.balancedAlignmentPlugin(context, dpn, log);
		
		// make confusion matrix
		int tp = 0;
		int fp = 0;
		int fn = 0;
		int tn = 0;
		
		// go through aligned fitting test log
		if (resultFitting != null) {
			for(XTrace trace : resultFitting.getAlignedLog()) {
				for (String key : trace.getAttributes().keySet()) {
					String traceName = trace.getAttributes().get(key).toString();
					try {
						Alignment gamma = resultFitting.getAlignmentByTraceName(traceName);
						Float fitness = gamma.getFitness();
						if (fitness < 1) {
							fn++;
						}
						else {
							tp++;
						}
						// System.out.println("Trace " + traceName + " has fitness " + Float.toString(fitness));
					}
					catch(NullPointerException e) {
						System.out.println("Trace " + traceName + " is not aligned");
//						for(XEvent event : trace) {
//							for(String attr : event.getAttributes().keySet()) {
//								if (attr.contains("name"))
//									System.out.print(event.getAttributes().get(attr) + ",");
//							}
//						}
//						System.out.println("");
					}
				}
			}
//			// add alignments to workspace
//			String alignmentsFittingName = "Alignments test log fitting " + i + " fold " + fold;
//			context.getProvidedObjectManager().createProvidedObject(alignmentsFittingName, 
//					resultFitting, BalancedReplayResult.class , context);
		}
		// go through aligned non-fitting test log
		if (resultNonFitting != null) {
			for(XTrace trace : resultNonFitting.getAlignedLog()) {
				for (String key : trace.getAttributes().keySet()) {
					String traceName = trace.getAttributes().get(key).toString();
					try {
						Alignment gamma = resultNonFitting.getAlignmentByTraceName(traceName);
						Float fitness = gamma.getFitness();
						if (fitness < 1) {
							tn++;
						}
						else {
							fp++;
						}
						// System.out.println("Trace " + traceName + " has fitness " + Float.toString(fitness));
					}
					catch(NullPointerException e) {
						System.out.println("Trace " + traceName + " is not aligned");
//						for(XEvent event : trace) {
//							for(String attr : event.getAttributes().keySet()) {
//								if (attr.contains("name"))
//									System.out.print(event.getAttributes().get(attr) + ",");
//							}
//						}
//						System.out.println("");
					}
				}
			}
//			// add alignments to workspace
//			String alignmentsNonFittingName = "Alignments test log non-fitting " + i + " fold " + fold;
//			context.getProvidedObjectManager().createProvidedObject(alignmentsNonFittingName, 
//					resultNonFitting, BalancedReplayResult.class , context);
		}
		
		// were all alignment computed?
		boolean allAlignments = ((testLogFitting.size() + testLogNonFitting.size()) == tp + fp + fn + tn);
		
		return new EvaluationResults(tp,fp,fn,tn,i,fold,allAlignments);
	}
	
	private static BalancedReplayResult replay(PluginContext context, BalancedProcessorConfiguration config, XLog log,
			DataPetriNet dpn, Marking initialMarking, Marking[] finalMarking,
			int index, int fold) throws DataAlignmentException {
		
		BalancedDataConformancePlusPlugin balanced = new BalancedDataConformancePlusPlugin();
		
		XEventClasses eventClasses = XUtils.createEventClasses(log.getClassifiers().get(0),log);
		Set<String> attributeNames = getAttributeNames(log, eventClasses.getClassifier());
		
		TransEvClassMapping activityMapping = BalancedProcessorConfiguration.createDefaultMappingTransitionsToEventClasses(dpn,
				eventClasses.getClassifier(), eventClasses);
		Map<XEventClass, Integer> mapEvClass2Cost = BalancedProcessorConfiguration.createDefaultLogMoveCost(eventClasses, 1);
		Map<Transition, Integer> mapTrans2Cost = BalancedProcessorConfiguration.createDefaultModelMoveCost(dpn, 1);
		
		Set<String> variableNames;
		variableNames = PetriNetWithData.getAllVariableNames(dpn);
		
		Map<String, String> variableMapping;
		if (!variableNames.isEmpty()) {
			variableMapping = BalancedProcessorConfiguration.createDefaultVariableMapping(variableNames, attributeNames);
		} else {
			variableMapping = ImmutableMap.of();
		}

		VariableMatchCosts variableCost;
		variableCost = BalancedProcessorConfiguration.createDefaultVariableCost(dpn, variableNames, 1, 1);
		
		config.setActivityMapping(activityMapping);
		config.setMapEvClass2Cost(mapEvClass2Cost);
		config.setMapTrans2Cost(mapTrans2Cost);
		config.setInitialMarking(initialMarking);
		// check final marking
		config.setFinalMarkings(finalMarking);
		config.setVariableCost(variableCost);
		config.setVariableMapping(variableMapping);
		// config.setObserver(new DataConformancePlusObserverImpl(context));
		final List<XTrace> unreliableTraceList = new ArrayList<XTrace>();
		config.setObserver(new DataConformancePlusObserverNoOpImpl() {

			public void unreliableAlignmentDetected(int traceIndex, XTrace currentTrace) {
				unreliableTraceList.add(currentTrace);
			}
		});
		
		BalancedProcessorConfiguration.autoGuessBounds(config, dpn, log);
		BalancedReplayResult result = null;
		try {
			result = balanced.doBalancedAlignmentDataConformanceChecking(dpn,
				log, context.getProgress(), config);
		}
		catch (ControlFlowAlignmentException e) {
			System.out.println("unreliable data alignment for " + index + " fold " + fold);
		}
		
		// add traces of impossible alignments to workspace
		if (unreliableTraceList.size() > 1) {
			XLog impossibleLog = addToLog(unreliableTraceList,log);
			String impossibleName = "Traces with impossible alignments " + index + " fold " + fold;
			context.getProvidedObjectManager().createProvidedObject(impossibleName, 
					impossibleLog, XLog.class , context);
		}
		return result;
	}
	
	private static XLog addToLog(List<XTrace> traces, XLog infoLog) {
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog output = factory.createLog((XAttributeMap) infoLog.getAttributes().clone());
		for(XTrace trace: traces) {
			output.add(trace);
		}
		return output;
	}
	
	private static Set<String> getAttributeNames(XLog log, XEventClassifier classifier) {
		Set<String> attributeNames;
		if (log.getInfo(classifier) != null) {
			attributeNames = ImmutableSet.copyOf(log.getInfo(classifier).getEventAttributeInfo().getAttributeKeys());
		} else {
			attributeNames = XUtils.getEventAttributeKeys(log);
		}
		return attributeNames;
	}
	
	/**
	* Cancel the future results, display a error message and return all nulls.
	* @param context Context of the plug-in.
	* @param msg Error message.
	* @return Return an object array of nulls;
	*/
	private EvaluationResultsCollection cancel(PluginContext context, NewickTreeCollection collection, String msg) {
		context.log(msg);
		context.getFutureResult(0).cancel(true);
		context.getFutureResult(1).cancel(true);
		return new EvaluationResultsCollection("default");
	}
}
