package org.processmining.ptandloggenerator.renderers;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.NumberFormat;
import java.util.Locale;

import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.table.DefaultTableModel;

import org.freehep.graphicsio.ps.EPSGraphics2D;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.LogarithmicAxis;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.NumberTickUnit;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYDotRenderer;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.processmining.contexts.uitopia.annotations.Visualizer;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.ProMSplitPane;
import org.processmining.framework.util.ui.widgets.ProMTable;
import org.processmining.ptandloggenerator.models.EvaluationResults;
import org.processmining.ptandloggenerator.models.EvaluationResultsCollection;

public class EvaluationResultsRenderer {
	
	private static final class StandardXYItemRenderer extends XYDotRenderer {
		private static final long serialVersionUID = 1018391957000784888L;

		public StandardXYItemRenderer() {
			super();
			this.setDotHeight(5);
			this.setDotWidth(5);
			setSeriesPaint(0, Color.BLACK);
			setSeriesPaint(1, Color.ORANGE);
		}
	}
	
	private static final class ChartListener implements MouseListener {
		private final double[] dataY;
		private final double[] dataX;
		private final ChartPanel panel;

		private ChartListener(double[] dataY, double[] dataX, ChartPanel panel) {
			this.dataY = dataY;
			this.dataX = dataX;
			this.panel = panel;
		}

		private ChartListener(double[] dataX, ChartPanel panel) {
			this.dataY = null;
			this.dataX = dataX;
			this.panel = panel;
		}

		public void mouseReleased(MouseEvent e) {
		}

		public void mousePressed(MouseEvent e) {
		}

		public void mouseExited(MouseEvent e) {
		}

		public void mouseEntered(MouseEvent e) {
		}

		public void mouseClicked(MouseEvent e) {
			if (e.getButton() == MouseEvent.BUTTON2) {
				final JFileChooser fc = new JFileChooser();
				fc.setFileSelectionMode(JFileChooser.FILES_ONLY);
				int retVal = fc.showSaveDialog(panel);
				if (retVal == JFileChooser.APPROVE_OPTION) {
					exportChart(fc.getSelectedFile(), panel, dataX, dataY);
				}
			}
		}
		
		private void exportChart(File file, ChartPanel panel, double[] dataX, double[] dataY) {
			export(new File(file.getAbsolutePath() + ".eps"), panel.getChart(), 400, 400);
			exportCSV(new File(file.getAbsolutePath() + ".csv"), dataX, dataY);
		}

		private void export(File file, JFreeChart chart, int x, int y) {
			try {
				EPSGraphics2D epsGraphics = new EPSGraphics2D(file, new Dimension(x, y));
				chart.draw(epsGraphics, new Rectangle(x, y));
				epsGraphics.closeStream();
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}
		
		protected void exportCSV(File selectedFile, double[] dataX, double[] dataY) {
			BufferedWriter bw = null;
			try {

				bw = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(selectedFile)));
				bw.write("\"dataX\"");
				if (dataY != null) {
					bw.write(",\"dataY\"");
				}

				final NumberFormat fNF = NumberFormat.getInstance(Locale.ENGLISH);
				fNF.setMaximumFractionDigits(5);
				fNF.setMinimumFractionDigits(0);

				for (int j = 0; j < dataX.length; j++) {
					bw.newLine();
					bw.write("\"" + fNF.format(dataX[j]) + "\"");
					if (dataY != null && j < dataY.length) {
						bw.write(",\"" + fNF.format(dataY[j]) + "\"");
					}
				}

			} catch (FileNotFoundException e) {
				throw new RuntimeException(e);
			} catch (IOException e) {
				throw new RuntimeException(e);
			} finally {
				if (bw != null) {
					try {
						bw.close();
					} catch (Exception exc) {
						// do nothing
					}
				}
			}
		}
	}
	
	@Plugin(name = "Simplified view on Decision Mining Evaluation Results", parameterLabels = { "Evaluation results collection" }, returnLabels = {
	"String Representation" }, returnTypes = { JComponent.class })
	@Visualizer
	
	@PluginVariant(requiredParameterLabels = { 0 })
	public static JComponent simpleVisualizer(PluginContext context, 
			EvaluationResultsCollection resultsCollection) {
		
		String [][] data = new String [resultsCollection.getResultsList().size()][];

		int i = 0;
		for (EvaluationResults results: resultsCollection.getResultsList()) {
			data[i] = new String [] {Integer.toString(results.getIndex()),
					Integer.toString(results.getFold()),
					Integer.toString(results.getTruePositives()),
					Integer.toString(results.getFalsePositives()),
					Integer.toString(results.getFalseNegatives()),
					Integer.toString(results.getTrueNegatives()),
					Boolean.toString(results.getAllAlignments()),
					resultsCollection.getTechnique()};
			i++;
		}
		
		DefaultTableModel tableModel = new DefaultTableModel(data, new String[] { "Index",
				"Fold", "True Positives", "False Positives",
				"False Negatives", "True Negatives", "All Alignments", "Techniques" });
		final ProMTable promTable = new ProMTable(tableModel);

		return promTable;		
	}
	
	@Plugin(name = "Visualize Decision Mining Evaluation Results", parameterLabels = { "Evaluation results collection" }, returnLabels = {
	"String Representation" }, returnTypes = { JComponent.class })
	@Visualizer
	
	@PluginVariant(requiredParameterLabels = { 0 })
	public static JComponent visualize(PluginContext context, EvaluationResultsCollection resultsCollection) {
		
		String [][] data = new String [resultsCollection.getResultsList().size()][];
		double [] recall = new double [resultsCollection.getResultsList().size()];
		double [] precision = new double [resultsCollection.getResultsList().size()];
		int i = 0;
		for (EvaluationResults results: resultsCollection.getResultsList()) {
			data[i] = new String [] {Integer.toString(results.getIndex()),
					Integer.toString(results.getFold()),
					Integer.toString(results.getTruePositives()),
					Integer.toString(results.getFalsePositives()),
					Integer.toString(results.getFalseNegatives()),
					Integer.toString(results.getTrueNegatives()),
					resultsCollection.getTechnique()};
			recall[i] = (double) results.getTruePositives() / (results.getFalseNegatives() + 
					results.getTruePositives());
			precision[i] = (double) results.getTruePositives() / (results.getFalsePositives() + 
					results.getTruePositives());
			i++;
		}
		
		ProMSplitPane parentPanel = new ProMSplitPane(ProMSplitPane.HORIZONTAL_SPLIT);
		
		DefaultTableModel tableModel = new DefaultTableModel(data, new String[] { "Index",
				"Fold", "True Positives", "False Positives",
				"False Negatives", "True Negatives", "Techniques" });
		final ProMTable promTable = new ProMTable(tableModel);
		
		JPanel chartPanel = addChart(recall, precision, "Scatterplot recall and precision",
				"recall", "precision", false, true);
		
		promTable.setPreferredSize(new Dimension(800, 400));
		parentPanel.setLeftComponent(promTable);
		parentPanel.setRightComponent(chartPanel);
		return parentPanel;		
	}
	
	private static ChartPanel addChart(double[] dataX, double[] dataY, String chartName, String xAxis, String yAxis,
			boolean showLogarithmicScale, boolean allowZero) {
		assert dataX.length == dataY.length : "Data X and Data Y does not match in size";
		XYSeries series = new XYSeries("default", false, true);
		series.setNotify(false);
		for (int i = 0; i < dataX.length; i++) {
			if (dataY[i] > 0 || allowZero) {
				series.add(dataX[i], dataY[i]);
			}
		}
		series.setNotify(true);
		XYDataset dataset = new XYSeriesCollection(series);
		JFreeChart chart = ChartFactory.createScatterPlot(chartName, xAxis, yAxis, dataset, PlotOrientation.VERTICAL,
				false, false, false);
		XYPlot plot = chart.getXYPlot();
		plot.setRenderer(new StandardXYItemRenderer());
		NumberAxis domain = (NumberAxis) plot.getDomainAxis();
		domain.setRange(-0.05, 1.05);
        domain.setTickUnit(new NumberTickUnit(0.1));
        domain.setVerticalTickLabels(true);
        NumberAxis range = (NumberAxis) plot.getRangeAxis();
        range.setRange(-0.05, 1.05);
        range.setTickUnit(new NumberTickUnit(0.1));
        
		if (showLogarithmicScale) {
			LogarithmicAxis logarithmicAxis = new LogarithmicAxis(yAxis);
			logarithmicAxis.setStrictValuesFlag(false);
			plot.setRangeAxis(logarithmicAxis);
		}
		ChartPanel panel = new ChartPanel(chart, false);
		panel.addMouseListener(new ChartListener(dataY, dataX, panel));
		return panel;
	}
	
	
}
